﻿using Couchbase;
using Couchbase.Configuration;
using Couchbase.Extensions;
using Enyim.Caching.Memcached;
using Newtonsoft.Json;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Concurrency
{
    class Program
    {
        #region Data Model
        public class Person
        {
            public string Name { get; set; }
            public double Age { get; set; }
            public Address Address { get; set; }
        }

        public class Address
        {
            public string City { get; set; }
            public string Street { get; set; }
            public int Number { get; set; }
        }
        #endregion

        private static CouchbaseClient _client;

        static void Main(string[] args)
        {
            var config = new CouchbaseClientConfiguration();
            config.Urls.Add(new Uri("http://localhost:8091/pools/"));
            config.Bucket = "default";

            _client = new CouchbaseClient(config);

            var person = new Person
            {
                Name = "Worker McThread",
                Age = 0,
                Address = null
            };

            var result = _client.StoreJson(StoreMode.Set, "_concurrencyTest", person);

            List<Task> tasks = new List<Task>();
            for (int i = 0; i < 10; i++)
                tasks.Add(Task.Run(() => 
                {
                    for (int j = 0; j < 1000; j++)
                    {
                        //UpdatePerson(p => {  p.Age++;  });
                       // UpdatePersonWithOptimisticConcurrency(p => {  p.Age++;  });
                        UpdatePersonWithPessimisticConcurrency(p => {  p.Age++;  });
                    }
                }));


            Task.WaitAll(tasks.ToArray());
            person = _client.GetJson<Person>("_concurrencyTest");
            Console.WriteLine("{0} is {1} years old.", person.Name, person.Age);
            Console.ReadLine();
        }


        private static void UpdatePerson(Action<Person> updateAction)
        {
                var person = _client.GetJson<Person>("_concurrencyTest");
                updateAction(person);
                var success = _client.StoreJson(StoreMode.Set, "_concurrencyTest", person);
        }

        private static void UpdatePersonWithOptimisticConcurrency(Action<Person> updateAction)
        {
            var success = false;
            var retries = 100;
            CasResult<string> result;
            Person person = null;

            do
            {
                result = _client.GetWithCas<string>("_concurrencyTest");
                person = JsonConvert.DeserializeObject<Person>(result.Result);
                updateAction(person);
                success = _client.CasJson(StoreMode.Replace, "_concurrencyTest", person, result.Cas);
            }
            while (!success && retries-- > 0);
        }

        private static void UpdatePersonWithPessimisticConcurrency(Action<Person> updateAction)
        {
            var success = false;
            var retries = 1000;
            CasResult<string> result;
            Person person = null;

            do
            {
                result = _client.GetWithLock<string>("_concurrencyTest", TimeSpan.FromSeconds(1));
                success = result.Result != null;
            }
            while (!success && retries-- > 0) ;

            person = JsonConvert.DeserializeObject<Person>(result.Result);
            updateAction(person);
            success = _client.CasJson(StoreMode.Set, "_concurrencyTest", person, result.Cas);
        }

    }
}
